# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Union, List, Tuple, Mapping

class DiscId:
    def __init__(self) -> None: ...
    def read(self, device: Optional[Union[str, bytes]], features: int = 0) -> None: ...
    def put(
        self, first: int, last: int, sectors: int, offsets: Union[List[int], Tuple[int]]
    ) -> None: ...
    @property
    def id(self) -> str: ...
    @property
    def freedb_id(self) -> str: ...
    @property
    def submission_url(self) -> str: ...
    @property
    def webservice_url(self) -> str: ...
    @property
    def first_track(self) -> int: ...
    @property
    def last_track(self) -> int: ...
    @property
    def sectors(self) -> int: ...
    @property
    def track_offsets(self) -> Tuple[int]: ...
    @property
    def track_lengths(self) -> Tuple[int]: ...
    @property
    def mcn(self) -> str: ...
    @property
    def track_isrcs(self) -> Tuple[str]: ...
    @property
    def device(self) -> Optional[str]: ...
    @property
    def toc(self) -> Optional[str]: ...

def default_device() -> str: ...

FEATURES_MAPPING: Mapping[int, str]
FEATURES: Tuple[str]
FEATURE_READ: int
FEATURE_MCN: int
FEATURE_ISRC: int
__discid_version__: str



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the libdiscid module
"""


class PutSuccess:
    first = 1
    last = 15
    sectors = 258725
    seconds = 3450
    offsets = (
        150,
        17510,
        33275,
        45910,
        57805,
        78310,
        94650,
        109580,
        132010,
        149160,
        165115,
        177710,
        203325,
        215555,
        235590,
    )
    track_seconds = (
        231,
        210,
        168,
        159,
        273,
        218,
        199,
        299,
        229,
        213,
        168,
        342,
        163,
        267,
        308,
    )
    disc_id = "TqvKjMu7dMliSfmVEBtrL7sBSno-"
    freedb_id = "b60d770f"
    toc = " ".join(map(str, [first, last, sectors] + list(offsets)))


class _PutFail:
    sectors = 200
    offsets = (1, 2, 3, 4, 5, 6, 7)


class PutFail1(_PutFail):
    first = 13
    last = 1


class PutFail2(_PutFail):
    first = 0
    last = 10


class PutFail2_2(_PutFail):
    first = 100
    last = 200


class PutFail3(_PutFail):
    first = 0
    last = 0


class PutFail3_2(_PutFail):
    first = 1
    last = 100



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""python-discid compat layer

This module provides a compatible layer so that python-libdiscid can be used as
a replacement for python-discid. It provides an interface compatible with
python-discid version 1.0.2.
"""

import libdiscid
import operator
import functools
import sys

_INVERSE_FEATURES = {
    libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ]: libdiscid.FEATURE_READ,
    libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_MCN]: libdiscid.FEATURE_MCN,
    libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_ISRC]: libdiscid.FEATURE_ISRC,
}


class _NoneHelper:
    def __getattr__(self, name):
        if name in (
            "id",
            "freedb_id",
            "submission_url",
            "toc",
            "first_track",
            "last_track",
            "sectors",
            "mcn",
        ):
            return None

        return super().__getattr__(name)


def _decode(string, encoding=None):
    # Let's do the same thing discid is doing. It always accepts both strings and
    # bytes objects and encodes/decodes them as it sees fit. libdiscid always
    # wants string objects, so let's decode it here on a best effort basis.
    if not isinstance(string, str):
        if encoding is None:
            encoding = sys.getfilesystemencoding() or "ascii"
        return string.decode(encoding)
    return string


# exceptions defined in discid
DiscError = libdiscid.DiscError


class TOCError(Exception):
    pass


# classes defined in discid
class Track:
    def __init__(self, disc, number):
        self.disc = disc
        self.number = number

    def __str__(self):
        return str(self.number)

    @property
    def offset(self):
        return self.disc.track_offsets[self.number - self.disc.first_track]

    @property
    def sectors(self):
        return self.disc.track_lengths[self.number - self.disc.first_track]

    length = sectors

    @property
    def seconds(self):
        return libdiscid.sectors_to_seconds(self.sectors)

    @property
    def isrc(self):
        try:
            value = self.disc.track_isrcs[self.number - self.disc.first_track]
        except NotImplementedError:
            return None
        return value if value != "" else None


class Disc:
    def __init__(self):
        self.disc = _NoneHelper()
        self.tracks = []

    def read(self, device, features=[]):
        self.disc = libdiscid.read(
            device,
            functools.reduce(
                operator.or_,
                (
                    _INVERSE_FEATURES[feature]
                    for feature in features
                    if feature in FEATURES
                ),
                0,
            ),
        )
        self.tracks = [
            Track(self.disc, numb)
            for numb in range(self.disc.first_track, self.disc.last_track + 1)
        ]
        return True

    def put(self, first, last, disc_sectors, track_offsets):
        try:
            self.disc = libdiscid.put(first, last, disc_sectors, track_offsets)
        except DiscError as disc_error:
            raise TOCError(str(disc_error))

        self.tracks = [
            Track(self.disc, num)
            for num in range(self.disc.first_track, self.disc.last_track + 1)
        ]
        return True

    @property
    def id(self):
        return self.disc.id

    @property
    def freedb_id(self):
        return self.disc.freedb_id

    @property
    def submission_url(self):
        return self.disc.submission_url

    @property
    def toc_string(self):
        return self.disc.toc

    @property
    def first_track_num(self):
        return self.disc.first_track

    @property
    def last_track_num(self):
        return self.disc.last_track

    @property
    def sectors(self):
        return self.disc.sectors

    length = sectors

    @property
    def seconds(self):
        return (
            libdiscid.sectors_to_seconds(self.sectors)
            if self.sectors is not None
            else None
        )

    @property
    def mcn(self):
        try:
            value = self.disc.mcn
        except NotImplementedError:
            return None
        return value if value != "" else None


# functions defined in discid
get_default_device = libdiscid.default_device


def read(device=None, features=[]):
    disc = Disc()
    disc.read(
        _decode(device) if device is not None else None,
        map(lambda feature: _decode(feature, "ascii"), features),
    )
    return disc


def put(first, last, disc_sectors, track_offsets):
    disc = Disc()
    disc.put(first, last, disc_sectors, track_offsets)
    return disc


# constants defined in discid
__version__ = "1.1.0 (compat layer from python-libdiscid %s)" % (libdiscid.__version__,)
"""This is the version of python-discid this layer is compatible with. """

LIBDISCID_VERSION_STRING = libdiscid.__discid_version__
FEATURES = libdiscid.FEATURES
FEATURES_IMPLEMENTED = (
    libdiscid.FEATURE_READ,
    libdiscid.FEATURE_MCN,
    libdiscid.FEATURE_ISRC,
)

__all__ = (
    "read",
    "put",
    "get_default_device",
    "__version__",
    "LIBDISCID_VERSION_STRING",
    "FEATURES",
    "FEATURES_IMPLEMENTED",
    "Disc",
    "DiscError",
    "TOCError",
)



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the libdiscid module
"""

import unittest
import libdiscid
from libdiscid import DiscError
from . import common


class TestLibDiscId(unittest.TestCase):
    def test_version(self):
        self.assertIsNotNone(libdiscid.__version__)
        self.assertIsNotNone(libdiscid.__discid_version__)

    def test_default_device(self):
        self.assertIsNotNone(libdiscid.DEFAULT_DEVICE)

    def test_default_device_2(self):
        self.assertIsNotNone(libdiscid.default_device())

    def test_features(self):
        self.assertIsNotNone(libdiscid.FEATURES)
        self.assertIsNotNone(libdiscid.FEATURE_READ)
        self.assertIsNotNone(libdiscid.FEATURE_MCN)
        self.assertIsNotNone(libdiscid.FEATURE_ISRC)
        self.assertIsNotNone(libdiscid.FEATURES_MAPPING)

    @unittest.skipIf(
        libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ] not in libdiscid.FEATURES,
        "not available on this platform",
    )
    def test_read_fail(self):
        self.assertRaises(DiscError, libdiscid.read, "/does/not/exist")

    def test_read_None(self):
        try:
            libdiscid.read()
        except (DiscError, NotImplementedError):
            pass

    @unittest.skipIf(
        libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ] in libdiscid.FEATURES,
        "available on this platform",
    )
    def test_read_not_implemented(self):
        self.assertRaises(NotImplementedError, libdiscid.read)

    def test_put(self):
        testdata = common.PutSuccess

        disc = libdiscid.put(
            testdata.first, testdata.last, testdata.sectors, testdata.offsets
        )
        self.assertIsNotNone(disc)
        self.assertIsNone(disc.device)

        self.assertEqual(disc.id, testdata.disc_id)
        self.assertEqual(disc.freedb_id, testdata.freedb_id)
        self.assertIsNotNone(disc.submission_url)
        self.assertEqual(disc.first_track, testdata.first)
        self.assertEqual(disc.last_track, testdata.last)
        self.assertEqual(disc.sectors, testdata.sectors)
        self.assertEqual(disc.pregap, testdata.offsets[0])
        self.assertEqual(disc.leadout_track, testdata.sectors)
        self.assertEqual(disc.toc, testdata.toc)

        self.assertEqual(len(disc.track_offsets), len(testdata.offsets))
        for read_offset, expected_offset in zip(disc.track_offsets, testdata.offsets):
            self.assertEqual(read_offset, expected_offset)

        # ISRCs are not available if one calls put
        if libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_ISRC] in libdiscid.FEATURES:
            self.assertEqual(len(disc.track_isrcs), 15)
            for read_isrc in disc.track_isrcs:
                self.assertEqual(read_isrc, "")

        # MCN is not available if one calls put
        if libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_MCN] in libdiscid.FEATURES:
            self.assertEqual(disc.mcn, "")

    def test_put_fail_1(self):
        # !(first < last)
        testdata = common.PutFail1
        self.assertRaises(
            DiscError,
            libdiscid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

    def test_put_fail_2(self):
        # !(first >= 1)
        testdata = common.PutFail2
        self.assertRaises(
            DiscError,
            libdiscid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

        # !(first < 100)
        testdata = common.PutFail2_2
        self.assertRaises(
            DiscError,
            libdiscid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

    def test_put_fail_3(self):
        # !(last >= 1)
        testdata = common.PutFail3
        self.assertRaises(
            DiscError,
            libdiscid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

        # !(last < 100)
        testdata = common.PutFail3_2
        self.assertRaises(
            DiscError,
            libdiscid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

    def test_sectors_to_seconds(self):
        # there are 75 sectors per second
        for sectors in range(38):
            # round down for sectors in the first half
            self.assertEqual(libdiscid.sectors_to_seconds(sectors), 0)
        for sectors in range(38, 76):
            # round up for sectors in the second half
            self.assertEqual(libdiscid.sectors_to_seconds(sectors), 1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the libdiscid.compat.discid module
"""

import unittest
import libdiscid
from libdiscid.compat import discid
from libdiscid.compat.discid import DiscError, TOCError
from . import common


class TestCompatDiscID(unittest.TestCase):
    def test_default_device(self):
        self.assertIsNotNone(discid.get_default_device())

    def test_features(self):
        self.assertIsNotNone(discid.FEATURES)

    def test_features_implementes(self):
        self.assertIsNotNone(discid.FEATURES_IMPLEMENTED)

    def test_empty_is_none(self):
        disc = discid.Disc()
        self.assertIsNone(disc.id)
        self.assertIsNone(disc.freedb_id)
        self.assertIsNone(disc.submission_url)
        self.assertIsNone(disc.toc_string)
        self.assertIsNone(disc.first_track_num)
        self.assertIsNone(disc.last_track_num)
        self.assertIsNone(disc.sectors)
        self.assertIsNone(disc.seconds)
        self.assertEqual(len(disc.tracks), 0)

    @unittest.skipIf(
        libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ] not in libdiscid.FEATURES,
        "not available on this platform",
    )
    def test_read_fail(self):
        self.assertRaises(DiscError, discid.read, "/does/not/exist")

    def test_read_None(self):
        try:
            discid.read()
        except (DiscError, NotImplementedError):
            pass

    @unittest.skipIf(
        libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ] not in libdiscid.FEATURES,
        "not available on this platform",
    )
    def test_encoded_device(self):
        self.assertRaises(DiscError, discid.read, "/does/not/exist")

    @unittest.skipIf(
        libdiscid.FEATURES_MAPPING[libdiscid.FEATURE_READ] not in libdiscid.FEATURES,
        "not available on this platform",
    )
    def test_byte_device(self):
        self.assertRaises(DiscError, discid.read, b"/does/not/exist")

    def test_put(self):
        testdata = common.PutSuccess

        disc = discid.put(
            testdata.first, testdata.last, testdata.sectors, testdata.offsets
        )
        self.assertIsNotNone(disc)

        self.assertEqual(disc.id, testdata.disc_id)
        self.assertEqual(disc.freedb_id, testdata.freedb_id)
        self.assertIsNotNone(disc.submission_url)
        self.assertEqual(disc.toc_string, testdata.toc)
        self.assertEqual(disc.first_track_num, testdata.first)
        self.assertEqual(disc.last_track_num, testdata.last)
        self.assertEqual(disc.sectors, testdata.sectors)
        self.assertEqual(disc.seconds, testdata.seconds)

        self.assertEqual(len(disc.tracks), len(testdata.offsets))
        for track, offset, sec in zip(
            disc.tracks, testdata.offsets, testdata.track_seconds
        ):
            self.assertEqual(track.offset, offset)
            self.assertEqual(track.seconds, sec)

        # ISRCs are not available if one calls put
        for track in disc.tracks:
            self.assertIsNone(track.isrc)

        # MCN is not available if one calls put
        self.assertIsNone(disc.mcn)

    def test_put_fail_1(self):
        # !(first < last)
        testdata = common.PutFail1
        self.assertRaises(
            TOCError,
            discid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

    def test_put_fail_2(self):
        # !(first >= 1)
        testdata = common.PutFail2
        self.assertRaises(
            TOCError,
            discid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

        # !(first < 100)
        testdata = common.PutFail2_2
        self.assertRaises(
            TOCError,
            discid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

    def test_put_fail_3(self):
        # !(last >= 1)
        testdata = common.PutFail3
        self.assertRaises(
            TOCError,
            discid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )

        # !(last < 100)
        testdata = common.PutFail3_2
        self.assertRaises(
            TOCError,
            discid.put,
            testdata.first,
            testdata.last,
            testdata.sectors,
            testdata.offsets,
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libc.string cimport const_char

cdef extern from "discid-wrapper.h":
    ctypedef void* DiscId

    DiscId* discid_new()
    void discid_free(DiscId* d)
    int wrap_read_sparse(DiscId *d, const_char* device, unsigned int features)
    int discid_put(DiscId *d, int first, int last, int *offsets)
    char *discid_get_error_msg(DiscId *d)
    char *discid_get_id(DiscId *d)
    char *discid_get_freedb_id(DiscId *d)
    char *discid_get_submission_url(DiscId *d)
    char *discid_get_webservice_url(DiscId *d)
    char *discid_get_default_device()
    int discid_get_first_track_num(DiscId *d)
    int discid_get_last_track_num(DiscId *d)
    int discid_get_sectors(DiscId *d)
    int discid_get_track_offset(DiscId *d, int track_num)
    int discid_get_track_length(DiscId *d, int track_num)

    cdef enum discid_feature:
        DISCID_FEATURE_READ
        DISCID_FEATURE_MCN
        DISCID_FEATURE_ISRC

    char* DISCID_FEATURE_STR_READ
    char* DISCID_FEATURE_STR_MCN
    char* DISCID_FEATURE_STR_ISRC

    int wrap_has_feature(int feature)
    char* wrap_get_version_string()

    char* wrap_get_mcn(DiscId *d)
    char* wrap_get_track_isrc(DiscId *d, int track_num)

    char* wrap_get_toc(DiscId *d)




# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Python bindings for libdiscid

libdiscid is a library to calculate MusicBrainz Disc IDs.
This module provides Python-bindings for libdiscid.

>>> disc = libdiscid.read()
>>> disc.id is not None
True
"""


from . import _discid
from .exceptions import DiscError
from typing import Optional, Union, List, Tuple
import re
import warnings

__version__ = "2.0"
__author__ = "Sebastian Ramacher"
__license__ = "MIT"
__copyright__ = f"(C) 2013-2021 {__author__}"

DEFAULT_DEVICE = _discid.default_device()
"""The default device to use for :func:`DiscId.read` on this platform.

.. deprecated:: 0.2.0
   Please use :func:`default_device` instead.
"""

FEATURES = _discid.FEATURES
"""Tuple of all available features supported by libdiscid on this platform.
"""
FEATURE_READ = _discid.FEATURE_READ
"""Read the TOC of the disc to get the disc ID. This feature is always enabled.
"""
FEATURE_MCN = _discid.FEATURE_MCN
"""Read the Media Catalogue Number of the disc.
"""
FEATURE_ISRC = _discid.FEATURE_ISRC
"""Read :musicbrainz:`International Standard Recording Codes <ISRC>` of all the
tracks.
"""
FEATURES_MAPPING = _discid.FEATURES_MAPPING
"""Mapping between the constants representing a feature and their string
representation.
"""
__discid_version__ = _discid.__discid_version__
"""The version of the underlying libdiscid.
"""


class DiscId:
    """Disc information

    Class holding all the information obtained from a disc.
    """

    def __init__(self, cdiscid: _discid.DiscId) -> None:
        self._id = cdiscid.id
        self._freedb_id = cdiscid.freedb_id
        self._submission_url = cdiscid.submission_url
        self._webservice_url = cdiscid.webservice_url
        self._first_track = cdiscid.first_track
        self._last_track = cdiscid.last_track
        self._sectors = cdiscid.sectors
        self._track_offsets = cdiscid.track_offsets
        self._track_lengths = cdiscid.track_lengths
        self._mcn = cdiscid.mcn
        self._track_isrcs = cdiscid.track_isrcs
        self._device = cdiscid.device
        self._toc = cdiscid.toc

    @property
    def id(self) -> str:
        """The MusicBrainz :musicbrainz:`Disc ID`."""

        return self._id

    @property
    def freedb_id(self) -> str:
        """The :musicbrainz:`FreeDB` Disc ID (without category)."""

        return self._freedb_id

    @property
    def submission_url(self) -> str:
        """Disc ID / TOC Submission URL for MusicBrainz

        With this url you can submit the current TOC as a new MusicBrainz
        :musicbrainz:`Disc ID`.
        """

        return self._submission_url

    @property
    def webservice_url(self) -> str:
        """The web service URL for info about the CD

        With this url you can retrieve information about the CD in XML from the
        MusicBrainz web service.
        """

        warnings.warn(
            "webservice_url is deprecated since it points to the old "
            "webservice. Please use python-musicbrainz-ngs to access "
            "the webservice.",
            DeprecationWarning,
        )
        return self._webservice_url

    @property
    def first_track(self) -> int:
        """Number of the first audio track."""

        return self._first_track

    @property
    def last_track(self) -> int:
        """Number of the last audio track."""

        return self._last_track

    @property
    def sectors(self) -> int:
        """Total sector count."""

        return self._sectors

    @property
    def leadout_track(self) -> int:
        """Leadout track."""

        return self.sectors

    @property
    def track_offsets(self) -> Tuple[int]:
        """Tuple of all track offsets (in sectors).

        The first element corresponds to the offset of the track denoted by
        :attr:`first_track` and so on.
        """

        return self._track_offsets

    @property
    def pregap(self) -> int:
        """Pregap of the first track (in sectors)."""

        return self.track_offsets[0]

    @property
    def track_lengths(self) -> Tuple[int]:
        """Tuple of all track lengths (in sectors).

        The first element corresponds to the length of the track denoted by
        :attr:`first_track` and so on.
        """

        return self._track_lengths

    @property
    def mcn(self) -> str:
        """Media Catalogue Number of the disc.

        :raises NotImplementedError: reading MCN is not supported on this platform
        """

        if self._mcn is None:
            raise NotImplementedError(
                "MCN is not available with this version " "of libdiscid and/or platform"
            )
        return self._mcn

    @property
    def track_isrcs(self) -> Tuple[str]:
        """Tuple of :musicbrainz:`ISRCs <ISRC>` of all tracks.

        The first element of the list corresponds to the ISRC of the
        :attr:`first_track` and so on.

        :raises NotImplementedError: reading ISRCs is not supported on this platform
        """

        if self._track_isrcs is None:
            raise NotImplementedError(
                "ISRC is not available with this version "
                "of libdiscid and/or platform"
            )
        return self._track_isrcs

    @property
    def device(self) -> Optional[str]:
        """The device the data was read from.

        If it is ``None``, :func:`libdiscid.put` was called to create the instance.
        """

        return self._device

    @property
    def toc(self) -> str:
        """String representing the CD's Table of Contents (TOC).

        :raises ValueError: extracting TOC string from the submission URL failed
        """

        if self._toc is None:
            # extract TOC string from submission URL
            match = re.match(r".*toc=([0-9+]+)$", self.submission_url)
            if match is None:
                raise ValueError("Failed to extract TOC from submission URL")
            self._toc = match.group(1).replace("+", " ")
        return self._toc


def read(device: Optional[Union[str, bytes]] = None, features: Optional[int] = None):
    """Reads the TOC from the device given as string.

    If *device* is ``None``, :func:`default_device` is used to determine
    the device. *features* can be any combination of :data:`FEATURE_MCN` and
    :data:`FEATURE_ISRC` and :data:`FEATURE_READ`. Note that prior to libdiscid
    version 0.5.0 *features* has no effect and that :data:`FEATURE_READ` is always
    assumed, even if not given.

    :param device: device to read from
    :type device: unicode or None
    :param features: selected features, possible values are :data:`FEATURE_READ` \
      :data:`FEATURE_MCN`, :data:`FEATURE_ISRC` and any of these values combined \
      with bitwise or.
    :type features: integer or None
    :raises libdiscid.DiscError: reading the disc failed
    :raises NotImplementedError: reading discs is not supported
    :raises MemoryError: failed to allocate the internal DiscId object
    :rtype: :class:`DiscId` object
    """

    disc = _discid.DiscId()
    if features is None:
        disc.read(device)
    else:
        disc.read(device, features)
    return DiscId(disc)


def put(
    first: int, last, sectors: int, offsets: Union[List[int], Tuple[int]]
) -> DiscId:
    """Creates a TOC based on the given offsets.

    Takes the *first* and *last* audio track, as well as the number of
    *sectors* and a list of *offsets* as in :attr:`track_offsets`.

    :param first: number of the first audio track
    :type first: integer
    :param last: number of the last audio track
    :type last: integer
    :param sectors: total number of sectors on the disc
    :type sectors: integer
    :param offsets: offsets of each track
    :type offsets: list or tuple of integers
    :raises libdiscid.DiscError: operation failed for some reason
    :raises MemoryError: failed to allocated memory to store the offsets or the \
      internal DiscId object
    :rtype: :class:`DiscId` object
    """

    disc = _discid.DiscId()
    disc.put(first, last, sectors, offsets)
    return DiscId(disc)


def default_device() -> str:
    """The default device on this platform.

    The default device can change during the run-time of the program. This can
    happen with removable devices for example.

    :rtype: str
    """

    return _discid.default_device()


def sectors_to_seconds(sectors: int) -> int:
    """Convert sectors to seconds rounded to the nearest second.

    :param sectors: number of sectors
    :type sectors: integer
    :rtype: integer
    """

    SECTORS_PER_SECOND = 75
    remainder = sectors % SECTORS_PER_SECOND
    return sectors // SECTORS_PER_SECOND + (
        1 if remainder > SECTORS_PER_SECOND // 2 else 0
    )


__all__ = (
    "read",
    "put",
    "default_device",
    "sectors_to_seconds",
    "__version__",
    "__discid_version__",
    "FEATURES",
    "FEATURES_MAPPING",
    "FEATURE_READ",
    "FEATURE_MCN",
    "FEATURE_ISRC",
    "DEFAULT_DEVICE",
    "DiscId",
    "DiscError",
)



# Copyright 2013-2021 Sebastian Ramacher <sebastian+dev@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cython: language_level=3, warn.unused=True

from . cimport cdiscid
cimport cpython
from libc cimport limits
from libc.stdlib cimport malloc, free
from cpython cimport bool
from libdiscid.exceptions import DiscError


cdef bool _has_feature(int feature):
    return cdiscid.wrap_has_feature(feature) == 1


cdef str _to_str(char* s):
    return s.decode('UTF-8', 'strict')


cdef class DiscId:
    """Class to calculate MusicBrainz Disc IDs.

    >>> d = DiscId()
    >>> d.read()
    >>> d.id is not None
    True

    Note that all the properties are only set after an successful call to
    :func:`DiscId.read` or :func:`DiscId.put`.
    """

    cdef cdiscid.DiscId *_c_discid
    cdef bool _have_read
    cdef str _device

    def __cinit__(self):
        self._c_discid = cdiscid.discid_new()
        if self._c_discid is NULL:
            raise MemoryError('Failed to allocate DiscId object')

        self._have_read = False
        self._device = None

    def __dealloc__(self):
        if self._c_discid is not NULL:
            cdiscid.discid_free(self._c_discid)

    cdef _read(self, char* device, unsigned int features):
        if not _has_feature(cdiscid.DISCID_FEATURE_READ):
            raise NotImplementedError(
                "read is not available with this version of libdiscid and/or platform"
            )

        if not cdiscid.wrap_read_sparse(self._c_discid, device, features):
            raise DiscError(self._get_error_msg())
        self._have_read = True

    def read(self, str device=None, unsigned int features=limits.UINT_MAX):
        """Reads the TOC from the device given as string.

        If *device* is ``None``, :func:`libdiscid.default_device` is used.
        *features* can be any combination of :data:`FEATURE_MCN` and
        :data:`FEATURE_ISRC` and :data:`FEATURE_READ`. Note that prior to libdiscid
        version 0.5.0 *features* has no effect and that :data:`FEATURE_READ` is
        always assumed, even if not given.

        A :exc:`libdiscid.DiscError` exception is raised when reading fails, and
        :py:exc:`NotImplementedError` when libdiscid does not support reading discs
        on the current platform.
        """

        if device is None:
            device = default_device()

        py_byte_device = device.encode('UTF-8')
        cdef char* cdevice = py_byte_device
        self._read(cdevice, features)
        self._device = device

    cdef _put(self, int first, int last, int* offsets):
        if not cdiscid.discid_put(self._c_discid, first, last, offsets):
            raise DiscError(self._get_error_msg())
        self._device = None
        self._have_read = True

    def put(self, int first, int last, int sectors, offsets):
        """Creates a TOC based on the given offsets.

        Takes the *first* and *last* audio track, as well as the number of
        *sectors* and a list of *offsets* as in :attr:`DiscId.track_offsets`.

        If the operation fails for some reason, a :exc:`libdiscid.DiscError`
        exception is raised.
        """

        cdef int* coffsets = <int*> malloc((len(offsets) + 1) * sizeof(int))
        if coffsets is NULL:
            raise MemoryError('Failed to allocate memory to store offsets')

        try:
            coffsets[0] = sectors
            for (i, v) in enumerate(offsets):
                coffsets[i + 1] = v
            return self._put(first, last, coffsets)
        finally:
            free(coffsets)

    cdef str _get_error_msg(self):
        return _to_str(cdiscid.discid_get_error_msg(self._c_discid))

    @property
    def id(self):
        """The MusicBrainz :musicbrainz:`Disc ID`.
        """

        return _to_str(cdiscid.discid_get_id(self._c_discid))

    @property
    def freedb_id(self):
        """The :musicbrainz:`FreeDB` Disc ID (without category).
        """

        return _to_str(cdiscid.discid_get_freedb_id(self._c_discid))

    @property
    def submission_url(self):
        """Disc ID / TOC Submission URL for MusicBrainz

        With this url you can submit the current TOC as a new MusicBrainz
        :musicbrainz:`Disc ID`.
        """

        return _to_str(cdiscid.discid_get_submission_url(self._c_discid))

    @property
    def webservice_url(self):
        """The web service URL for info about the CD

        With this url you can retrieve information about the CD in XML from the
        MusicBrainz web service.
        """

        return _to_str(cdiscid.discid_get_webservice_url(self._c_discid))

    @property
    def first_track(self):
        """Number of the first audio track.
        """

        return cdiscid.discid_get_first_track_num(self._c_discid)

    @property
    def last_track(self):
        """Number of the last audio track.
        """

        return cdiscid.discid_get_last_track_num(self._c_discid)

    @property
    def sectors(self):
        """Total sector count.
        """

        return cdiscid.discid_get_sectors(self._c_discid)

    @property
    def track_offsets(self):
        """Tuple of all track offsets (in sectors).

        The first element corresponds to the offset of the track denoted by
        :attr:`first_track` and so on.
        """

        return tuple(
            cdiscid.discid_get_track_offset(self._c_discid, track)
            for track in range(self.first_track, self.last_track + 1)
        )

    @property
    def track_lengths(self):
        """Tuple of all track lengths (in sectors).

        The first element corresponds to the length of the track denoted by
        :attr:`first_track` and so on.
        """

        return tuple(
            cdiscid.discid_get_track_length(self._c_discid, track)
            for track in range(self.first_track, self.last_track + 1)
        )

    @property
    def mcn(self):
        """Media Catalogue Number of the disc.
        """

        if not _has_feature(cdiscid.DISCID_FEATURE_MCN):
            return None
        return _to_str(cdiscid.wrap_get_mcn(self._c_discid))

    @property
    def track_isrcs(self):
        """Tuple of :musicbrainz:`ISRCs <ISRC>` of all tracks.

        The first element of the list corresponds to the ISRC of the
        :attr:`first_track` and so on.
        """

        if not _has_feature(cdiscid.DISCID_FEATURE_ISRC):
          return None
        return tuple(
            _to_str(cdiscid.wrap_get_track_isrc(self._c_discid, track))
            for track in range(self.first_track, self.last_track + 1)
        )

    @property
    def device(self):
        """The device the data was read from.

        If it is ``None``, :func:`libdiscid.put` was called to create the instance.
        """

        return self._device

    @property
    def toc(self):
        """String representing the CD's Table of Contents (TOC).
        """

        assert self._have_read

        cdef char* tocstr = cdiscid.wrap_get_toc(self._c_discid)
        if tocstr is not NULL:
            return _to_str(tocstr)
        return None


FEATURES_MAPPING = {
    cdiscid.DISCID_FEATURE_READ: _to_str(cdiscid.DISCID_FEATURE_STR_READ),
    cdiscid.DISCID_FEATURE_MCN: _to_str(cdiscid.DISCID_FEATURE_STR_MCN),
    cdiscid.DISCID_FEATURE_ISRC: _to_str(cdiscid.DISCID_FEATURE_STR_ISRC)
}


cdef _feature_list():
    return tuple(s for f, s in FEATURES_MAPPING.items() if _has_feature(f))


def default_device():
    """The default device on this platform.
    """

    return _to_str(cdiscid.discid_get_default_device())


FEATURES = _feature_list()
FEATURE_READ = cdiscid.DISCID_FEATURE_READ
FEATURE_MCN = cdiscid.DISCID_FEATURE_MCN
FEATURE_ISRC = cdiscid.DISCID_FEATURE_ISRC
__discid_version__ = _to_str(cdiscid.wrap_get_version_string())



#!/usr/bin/python3

import os.path
import os
import sys

from setuptools import setup, Extension

try:
    import pkgconfig

    have_pkgconfig = True

    def pkgconfig_exists(package):
        try:
            return pkgconfig.exists(package)
        except OSError:
            return False


except ImportError:
    have_pkgconfig = False

if have_pkgconfig and pkgconfig_exists("libdiscid"):
    flags = pkgconfig.parse("libdiscid")
    define_macros = flags["define_macros"]
    include_dirs = flags["include_dirs"]
    library_dirs = flags["library_dirs"]
    libraries = list(flags["libraries"])
else:
    define_macros = []
    include_dirs = []
    library_dirs = []
    libraries = ["discid"]

    # TODO: Solve this properly by passing CFLAGS and LDFLAGS via CIBW_ENVIRONMENT. For some reason,
    # CFLAGS and LDFLAGS seem to be ignored when built through cibuildwheel.
    LIBDISCID_HOME = os.environ.get("LIBDISCID_HOME", None)
    if LIBDISCID_HOME is not None:
        library_dirs.append(os.path.join(LIBDISCID_HOME, "libdiscid-0.6.1-win32"))
        include_dirs.append(
            os.path.join(
                LIBDISCID_HOME, "libdiscid-0.6.1-win32", "libdiscid-0.6.1", "include"
            )
        )


setup(
    ext_modules=[
        Extension(
            "libdiscid._discid",
            ["libdiscid/_discid.pyx", "libdiscid/discid-wrapper.c"],
            define_macros=define_macros,
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=libraries,
        )
    ],
    packages=["libdiscid", "libdiscid.tests", "libdiscid.compat"],
    package_data={
        "libdiscid": ["_discid.pyi", "py.typed"],
    },
    test_suite="libdiscid.tests",
)
